"""
Module defining the TimeSeries object
"""

from typing import IO, Union
from pathlib import Path
import warnings

from dataclasses import dataclass
import numpy as np

from numpy.typing import NDArray


class ShapeChangedWarning(Warning):

    """
    warn the user that TimeSeries.__post_init__ changed the shape of the dependent variable
    """


@dataclass
class TimeSeries:

    dependent_variable: NDArray
    times: NDArray

    def __post_init__(self):

        timesteps = np.diff(self.times)
        if not np.isclose(np.std(timesteps), 0.0):
            raise ValueError("TimeSeries.times must be uniformly spaced")
        if np.isclose(np.mean(timesteps), 0.0):
            raise ValueError("TimeSeries.times must have a timestep greater than zero")

        if len(self.dependent_variable.shape) == 1:
            num_steps = len(self.dependent_variable)
            self.dependent_variable = self.dependent_variable.reshape(num_steps, 1)
            warnings.warn(
                f"TimeSeries.dependent_variable should have shape (number of steps, dimensionality). "
                f"The shape has been changed from {(num_steps,)} to {self.dependent_variable.shape}"
            )

    def save(self, fp: Union[IO, str, Path], header: str = "", delimiter=","):

        """
        Method to save a time series

        Args:
            fp (Union[IO, str, Path]):
                The file-like object, path name, or Path in which to save the TimeSeries instance

            header (str):
                An optional header. Defaults to the empty string

            delimiter (str):
                The delimiting character in the save file. Defaults to a comma

        """
        np.savetxt(
            fp,
            np.vstack((self.times, self.dependent_variable.T)).T,
            delimiter=delimiter,
            header=header,
            comments=""
        )

    @property
    def num_dims(self) -> int:

        """
        The dimensionality of the time series

        Returns:
            int: The dimensionality of the time series
        """

        return self.dependent_variable.shape[1]

    @classmethod
    def from_csv(cls, fp: Union[IO, str, Path], time_index: int = 0):

        """
        Method for loading in a TimeSeries instance from a comma-separated value (csv) file

        Args:
            fp (Union[IO, str, Path]):
                The file-like object, path name, or Path in which to read

            time_index (int):
                The column index corresponding to the time column. Defaults to 0

        Returns:
            TimeSeries: A TimeSeries instance populated by data from the csv file
        """

        data = np.loadtxt(fp, delimiter=",")

        return cls(
            dependent_variable=np.delete(data, obj=time_index, axis=1),
            times=data[:, time_index]
        )

    @property
    def timestep(self) -> float:

        """
        The physical timestep of the time series

        Returns:
            int: The physical timestep of the time series
        """

        return self.times[1] - self.times[0]

    def __eq__(self, other) -> bool:
        return bool(np.all(self.times == other.times) and np.all(
            np.isclose(self.dependent_variable, other.dependent_variable)
        ))

    def __getitem__(self, key):

        return TimeSeries(self.dependent_variable[key], self.times[key])

    def __setitem__(self, key, value):

        if not isinstance(value, TimeSeries):
            raise TypeError("Value must be a TimeSeries object")
        if isinstance(key, slice) and key.stop > len(self.dependent_variable):
            raise ValueError("Slice stop index out of range")
        if isinstance(key, int) and key >= len(self.dependent_variable):
            raise ValueError("Index out of range")

        self.dependent_variable[key] = value.dependent_variable
        self.times[key] = value.times

    def __add__(self, other):

        if not len(self.times) == len(other.times):
            raise ValueError("can only add TimeSeries instances that have the same number of timesteps")

        if not np.all(self.times == other.times):
            raise ValueError("can only add TimeSeries instances that span the same times")

        return TimeSeries(
            dependent_variable=self.dependent_variable + other.dependent_variable,
            times=self.times
        )

    def __sub__(self, other):

        if not len(self.times) == len(other.times):
            raise ValueError("can only subtract TimeSeries instances that have the same number of timesteps")

        if not np.all(self.times == other.times):
            raise ValueError("can only subtract TimeSeries instances that span the same times")

        return TimeSeries(
            dependent_variable=self.dependent_variable - other.dependent_variable,
            times=self.times
        )

    def __rshift__(self, other):

        if self.times[-1] >= other.times[0]:
            raise ValueError("can only concatenate TimeSeries instances with non-overlapping time values")

        return TimeSeries(
            dependent_variable=np.vstack((self.dependent_variable, other.dependent_variable)),
            times=np.hstack((self.times, other.times))
        )

    def __len__(self):

        return len(self.times)



"""
Module defining reservoirs
"""


from dataclasses import dataclass
from abc import ABC, abstractmethod
from typing import Callable, Optional

import numpy as np


@dataclass
class Reservoir(ABC):

    """
    Abstract base reservoir class defining input state -> reservoir state
    """

    input_dimensionality: int
    """dimensionality of the input state"""

    reservoir_dimensionality: int
    """dimensionality of the reservoir state, equivalently the reservoir size"""

    @abstractmethod
    def input_to_reservoir(self, input_state: np.typing.NDArray) -> np.typing.NDArray:

        """
        Map from input state to reservoir state

        Args:
            input_state: input state to map to reservoir state
        """

        pass


@dataclass
class DynamicalReservoir(Reservoir):

    """
    Dynamical reservoir state, defined by the mapping:
    y_t = f(w_in @ x_t)
    where f is some nonlinear activation function
    """

    generator: Optional[np.random.Generator] = None
    """
    random generator for the class to use
    will be set to np.random.default_rng(seed=0) if not specified
    """

    w_in: Optional[np.typing.NDArray] = None
    """
    input linear mapping. must be shape (self.reservoir_dimensionality, self.input_dimensionality)
    if not defined at initialization, will be auto generated
    """

    activation_function: Callable[[np.typing.NDArray], np.typing.NDArray] = np.tanh
    """
    activation function f. defaults to np.tanh
    """

    def __post_init__(self):

        if self.generator is None:
            self.generator = np.random.default_rng(seed=0)

        if self.w_in is None:
            self.w_in = self.generator.uniform(
                low=-0.5,
                high=0.5,
                size=(self.reservoir_dimensionality, self.input_dimensionality)
            )

    def input_to_reservoir(self, input_state: np.typing.NDArray) -> np.typing.NDArray:

        """
        Map from input state to reservoir state via y_t = f(w_in @ x_t)

        Args:
            input_state: input state to map to reservoir state
        """

        return self.activation_function(self.w_in @ input_state)



"""
Module defining models, i.e. compositions of reservoir(s) and readout layers
"""

from dataclasses import dataclass, field

import numpy as np

from .reservoirs import Reservoir
from .readouts import Readout
from .time_series import TimeSeries


@dataclass
class Model:

    """
    Model class for training and predicting
    """

    reservoir: Reservoir
    """reservoir defining input -> reservoir mapping"""

    readout: Readout
    """readout defining reservoir -> output mapping"""

    final_time: float = field(init=False)
    """final time read during training. will be set at training"""

    timestep: float = field(init=False)
    """timestep read during training. will be set at training"""

    initial_guess: np.typing.NDArray = field(init=False)
    """initial guess read during training. will be set at training"""

    def train(
        self,
        input_time_series: TimeSeries,
        warmup: int = 0,
        rcond: float = 1.0e-10
    ):
        """
        Training method for model

        Args:
            input_time_series: TimeSeries instance to train on
            warmup: Number of initial steps to ignore in training
            rcond: Threshold for pseudo-inverse calculation. Increase if prediction is unstable
        """

        if warmup >= len(input_time_series.times):
            raise ValueError(f"warmup must be smaller than number of timesteps ({len(input_time_series)})")

        time_series_array = input_time_series.dependent_variable
        independent_variables = self.reservoir.input_to_reservoir(time_series_array[:-1, :].T).T
        dependent_variables = time_series_array[1:]

        if warmup > 0:
            independent_variables = independent_variables[warmup:]
            dependent_variables = dependent_variables[warmup:]

        w_out_transpose = np.linalg.pinv(independent_variables, rcond=rcond) @ dependent_variables
        self.readout.coefficients = w_out_transpose.T
        self.timestep = input_time_series.timestep
        self.final_time = input_time_series.times[-1]
        self.initial_guess = time_series_array[-1, :]

    def predict(self, horizon: int) -> TimeSeries:

        """
        Model prediction method

        Args:
            horizon: number of steps to forecast into the future
        """

        predictions = np.zeros((horizon, self.reservoir.input_dimensionality))

        for i in range(horizon):
            if i == 0:
                predictions[i, :] = self.readout.reservoir_to_output(
                    self.reservoir.input_to_reservoir(self.initial_guess)
                )
                continue
            predictions[i, :] = self.readout.reservoir_to_output(
                self.reservoir.input_to_reservoir(predictions[i - 1, :])
            )

        return TimeSeries(
            dependent_variable=predictions,
            times=self.final_time + self.timestep + np.linspace(0, horizon * self.timestep, horizon)
        )



"""
Module defining readout layers
"""

from dataclasses import dataclass, field
from abc import ABC, abstractmethod

import numpy as np


@dataclass
class Readout(ABC):

    """
    Abstract readout layer, defining reservoir state -> output state mapping
    """

    coefficients: np.typing.NDArray = field(init=False)
    """coefficients defining the readout layer, which will later be fitted"""

    @abstractmethod
    def reservoir_to_output(self, reservoir_state: np.typing.NDArray) -> np.typing.NDArray:

        """
        Map from reservoir state to output

        Args:
            reservoir_state: reservoir state to map to output state
        """

        pass


@dataclass
class LinearReadout(Readout):

    """
    Linear readout layer, reservoir_state -> W @ reservoir_state
    """

    def reservoir_to_output(self, reservoir_state: np.typing.NDArray) -> np.typing.NDArray:

        """
        Linearly from reservoir state to output, i.e. just self.coefficients @ reservoir_state

        Args:
            reservoir_state (np.typing.NDArray):
                reservoir state to map to output state
        """

        if not hasattr(self, "coefficients"):
            raise ValueError("Need to train readout before using it")

        return self.coefficients @ reservoir_state
